from typing import Any, Dict, List, Tuple

class LowLevelBlock:
    def __init__(self, id: str, name: str, block_type: str, block_class: str, **kwargs):
        self.id = id
        self.name = name
        self.block_type = block_type
        self.block_class = block_class
        self.extra = kwargs  # Any additional block-specific properties

    def to_dict(self) -> Dict[str, Any]:
        d = {
            "Id": self.id,
            "Name": self.name,
            "BlockType": self.block_type,
            "BlockClass": self.block_class,
        }
        d.update(self.extra)
        return d

class LowLevelLink:
    def __init__(
        self,
        id: str,
        name: str,
        source_block_id: str,
        source_port_idx: int,
        destination_block_id: str,
        destination_port_idx: int,
    ):
        self.id = id
        self.name = name
        self.source_block_id = source_block_id
        self.source_port_idx = source_port_idx
        self.destination_block_id = destination_block_id
        self.destination_port_idx = destination_port_idx

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Id": self.id,
            "Name": self.name,
            "SourceBlockId": self.source_block_id,
            "SourcePortIdx": self.source_port_idx,
            "DestinationBlockId": self.destination_block_id,
            "DestinationPortIdx": self.destination_port_idx,
        }

class LowLevelBlockStructure:
    def __init__(
        self,
        blocks: List[LowLevelBlock],
        links: List[LowLevelLink],
        port_map: Dict[Tuple[str, int], Tuple[str, int]],
    ):
        """
        port_map: maps (high_level_port_type, port_index) -> (low_level_block_id, low_level_port_index)
        Example: {("input", 0): ("sum1", 1), ("output", 0): ("display1", 0)}
        """
        self.blocks = blocks
        self.links = links
        self.port_map = port_map

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Blocks": [block.to_dict() for block in self.blocks],
            "Links": [link.to_dict() for link in self.links],
            "PortMap": {
                f"{ptype}_{pidx}": {"block_id": bid, "port_idx": lidx}
                for (ptype, pidx), (bid, lidx) in self.port_map.items()
            },
        }


import json
import os
import runpy
import yaml
import pathlib
from typing import Dict, Any, List
from pysyslink_toolkit.Plugin import Plugin
from pysyslink_toolkit.HighLevelBlock import HighLevelBlock
from pysyslink_toolkit.LowLevelBlockStructure import LowLevelBlock, LowLevelLink, LowLevelBlockStructure
from pysyslink_toolkit.load_plugins import load_plugins_from_paths

def compile_high_level_block(block: HighLevelBlock, plugins: list[Plugin]) -> LowLevelBlockStructure:
    for plugin in plugins:
        try:
            return plugin.compile_block(block)
        except NotImplementedError:
            continue
    raise RuntimeError(f"No plugin could compile block: {block.block_type}")

def format_property_value(prop_type, value):
    try:
        if prop_type == "int":
            return str(int(value))
        elif prop_type == "float":
            return f"{float(value):.1f}"
        elif prop_type == "double":
            return f"{float(value):.1f}"
        elif prop_type == "float[]":
            return [f"{float(v):.1f}" for v in value]
        elif prop_type == "double[]":
            return [f"{float(v):.1f}" for v in value]
        elif prop_type == "int[]":
            return [str(int(v)) for v in value]
        else:
            return value
    except Exception as e:
        raise ValueError(f"Error while formatting property type {prop_type} value {value}: {e}")

def serialize_block(block: LowLevelBlock) -> dict:
    d = block.to_dict()
    if "properties" in d:
        for k, v in d["properties"].items():
            if isinstance(v, dict) and "type" in v and "value" in v:
                d["properties"][k]["value"] = format_property_value(v["type"], v["value"])
    return d

def compile_pslk_to_yaml(pslk_path: str, config_path: str, output_yaml_path: str):
    # Load the .pslk file (JSON)
    with open(pslk_path, "r") as f:
        system_json = json.load(f)

    # Load plugins
    config = yaml.safe_load(open(config_path))
    plugin_paths = config.get("plugin_paths", [])
    plugins = load_plugins_from_paths(config_path, plugin_paths)

    print("All plugins obtained")

    initialization_python_script_path = system_json.get("initialization_python_script_path", None)


    if initialization_python_script_path:
        # Resolve to absolute path if not already absolute
        if not os.path.isabs(initialization_python_script_path):
            pslk_dir = os.path.dirname(os.path.abspath(pslk_path))
            initialization_python_script_path = os.path.normpath(
                os.path.join(pslk_dir, initialization_python_script_path)
            )
        if os.path.isfile(initialization_python_script_path) and initialization_python_script_path.endswith(".py"):
            try:
                parameter_environment_dict = runpy.run_path(initialization_python_script_path, init_globals={})
            except Exception as e:
                raise RuntimeError(f"Initialization script {initialization_python_script_path} load failed") from e
        else:
            raise FileNotFoundError(f"Initialization script '{initialization_python_script_path}' not found or not a .py file.")
    else:
        print(f"No initialization script provided.")
        parameter_environment_dict = dict()


    # Compile each high-level block
    block_structs: Dict[str, LowLevelBlockStructure] = {}
    for block_data in system_json.get("blocks", []):
        block = HighLevelBlock.from_dict(block_data, parameter_environment_dict)
        ll_struct = compile_high_level_block(block, plugins)
        block_structs[block.id] = ll_struct

    print("High level blocks compiled")

    # Collect all low-level blocks and links
    all_blocks: List[LowLevelBlock] = []
    all_links: List[LowLevelLink] = []
    port_maps: Dict[str, Dict[str, Any]] = {}

    for block_id, struct in block_structs.items():
        all_blocks.extend(struct.blocks)
        all_links.extend(struct.links)
        port_maps[block_id] = struct.port_map

    # Now resolve high-level links to low-level links using port maps
    link_idx = 1
    for link in system_json.get("links", []):
        src_id = link["sourceId"]
        src_port = link["sourcePort"]
        tgt_id = link["targetId"]
        tgt_port = link["targetPort"]

        src_map = port_maps.get(src_id, {})
        tgt_map = port_maps.get(tgt_id, {})
        src_ll = src_map.get(("output", src_port))
        tgt_ll = tgt_map.get(("input", tgt_port))
        if not src_ll or not tgt_ll:
            raise RuntimeError(f"Cannot resolve link: {link}")

        src_block_id, src_port_idx = src_ll
        tgt_block_id, tgt_port_idx = tgt_ll

        ll_link = LowLevelLink(
            id=link.get("id", f"link{link_idx}"),
            name=link.get("id", f"link{link_idx}"),
            source_block_id=src_block_id,
            source_port_idx=src_port_idx,
            destination_block_id=tgt_block_id,
            destination_port_idx=tgt_port_idx,
        )
        all_links.append(ll_link)
        link_idx += 1

    # Prepare YAML output with formatted properties
    output = {
        "Blocks": [serialize_block(block) for block in all_blocks],
        "Links": [link.to_dict() for link in all_links],
    }

    with open(output_yaml_path, "w") as f:
        yaml.dump(output, f, sort_keys=False)

# Example usage:
# compile_pslk_to_yaml("test_json.pslk", "toolkit_config.yaml", "output_system.yaml")


import os
import pytest
import yaml
from pysyslink_toolkit import api
import asyncio

TEST_DIR = os.path.dirname(__file__)
OUTPUT_DIR = os.path.join(os.path.dirname(__file__), "test_outputs")
os.makedirs(OUTPUT_DIR, exist_ok=True)

@pytest.fixture
def config_path():
    return os.path.join(TEST_DIR, "data", "toolkit_config.yaml")

@pytest.fixture
def pslk_path():
    return os.path.join(TEST_DIR, "data", "dummy.pslk")

@pytest.fixture
def pslk_path_with_parameters():
    return os.path.join(TEST_DIR, "data", "parse_parameters.pslk")

@pytest.fixture
def output_yaml_path(request):
    OUTPUT_DIR = os.path.join(TEST_DIR, "test_outputs")
    os.makedirs(OUTPUT_DIR, exist_ok=True)
    test_name = request.node.name  # This gives you the test function name
    return os.path.join(OUTPUT_DIR, f"output_{test_name}.yaml")

def test_compile_system(config_path, pslk_path, output_yaml_path):
    result = api.compile_system(config_path, pslk_path, str(output_yaml_path))
    assert result == "success"
    with open(output_yaml_path) as f:
        content = f.read()
        assert "Blocks:" in content

def test_compile_system_with_parameters(config_path, pslk_path_with_parameters, output_yaml_path):
    result = api.compile_system(config_path, pslk_path_with_parameters, str(output_yaml_path))
    assert result == "success"
    with open(output_yaml_path) as f:
        content = f.read()
        assert "Blocks:" in content

def test_neuron_block_compilation(config_path, output_yaml_path):
    # Paths to input and expected output
    test_dir = os.path.dirname(__file__)
    input_pslk = os.path.join(test_dir, "data", "neuron_test.pslk")
    expected_yaml = os.path.join(test_dir, "data", "neuron_output.yaml")
    output_yaml = output_yaml_path

    # Run the compilation
    result = api.compile_system(config_path, input_pslk, str(output_yaml))
    assert result == "success"

    # Load and compare YAMLs
    with open(output_yaml) as f:
        actual = yaml.safe_load(f)
    with open(expected_yaml) as f:
        expected = yaml.safe_load(f)

    assert actual == expected

def test_get_available_block_libraries(config_path):
    libs = api.get_available_block_libraries(config_path)
    assert isinstance(libs, list)
    assert any(lib.name == "dummy_library" for lib in libs)

def test_get_block_render_information(config_path):
    block_data = {
        "id": "block1",
        "label": "Test Block",
        "inputPorts": 1,
        "outputPorts": 1,
        "blockLibrary": "dummy_library",
        "blockType": "dummy",
        "properties": {}
    }
    test_dir = os.path.dirname(__file__)
    input_pslk = os.path.join(test_dir, "data", "neuron_test.pslk")

    info = api.get_block_render_information(config_path, block_data, input_pslk)
    assert info is not None

def test_get_block_render_information_with_parameters(config_path):
    block_data = {
        "id": "nx0MITpddR4PS825bmSZzPmHFqtCAu6b",
        "blockLibrary": "core_BasicBlocks",
        "blockType": "Constant",
        "label": "Constant",
        "x": 707.0833333333333,
        "y": 518.5538194444446,
        "inputPorts": 0,
        "outputPorts": 1,
        "properties": {
            "Value": {
            "type": "float",
            "value": "max(4, 5) - 4/3"
            }
        }
    }
    test_dir = os.path.dirname(__file__)
    input_pslk = os.path.join(test_dir, "data", "neuron_test.pslk")
    info = api.get_block_render_information(config_path, block_data, input_pslk)
    assert info is not None

def test_get_block_render_information_basic_block(config_path):
    block_data = {
        "id":"InIvNfx88BzB6k7e7uL2ODxhMcVDWbH0",
        "blockLibrary":"core_BasicBlocks",
        "blockType":"Constant",
        "label":"Constant",
        "x":90,
        "y":146,
        "inputPorts":1,
        "outputPorts":1,"properties":{
            "Value": {
                "type": "float",
                "value": 1
            }
        }
    }
    test_dir = os.path.dirname(__file__)
    input_pslk = os.path.join(test_dir, "data", "neuron_test.pslk")
    info = api.get_block_render_information(config_path, block_data, input_pslk)
    print(info.to_dict())
    assert info is not None
    assert info.input_ports == 0
    assert info.output_ports == 1
    assert info.text == "Constant"


def test_simulation_runs_and_callbacks(config_path):
    test_dir = os.path.dirname(__file__)
    system_yaml = os.path.join(test_dir, "data", "simulable_system.yaml")
    sim_options_yaml = os.path.join(test_dir, "data", "sim_options.yaml")
    output_json = os.path.join(OUTPUT_DIR, "output_test_simulation_runs_and_callbacks.json")

    # Prepare a callback to record display updates
    callback_calls = []

    def display_callback(event):
        callback_calls.append((event.value_id, event.simulation_time, event.value))

    # Run the simulation
    result = asyncio.run(
        api.run_simulation(
            config_path,
            system_yaml,
            sim_options_yaml,
            display_callback=display_callback
        )
    )

    assert os.path.exists(output_json)
    # The simulation result object should not be None
    assert result is not None
    # The callback should have been called at least once
    assert len(callback_calls) > 0


from typing import Any, Dict

def safe_eval(expr: str, context: Dict[str, Any] = {}) -> Any:
    """Evaluate a Python expression safely in a minimal context."""
    allowed_builtins = {
        'min': min,
        'max': max,
        'abs': abs,
        'sum': sum,
        'len': len,
        'float': float,
        'int': int,
        'range': range,
        # Add math functions if needed
    }
    import math
    safe_globals = {"__builtins__": allowed_builtins, "math": math}
    safe_globals.update(context)
    return eval(expr, safe_globals, {})

class HighLevelBlock:
    def __init__(
        self,
        id: str,
        label: str,
        input_ports: int,
        output_ports: int,
        block_library: str,
        block_type: str,
        properties: Dict[str, Dict[str, Any]]
    ):
        self.id = id
        self.label = label
        self.input_ports = input_ports
        self.output_ports = output_ports
        self.block_library = block_library
        self.block_type = block_type
        self.properties = properties

    @classmethod
    def from_dict(cls, data: Dict[str, Any], parameter_environment_namespace: Dict[str, Any]) -> "HighLevelBlock":
        required_fields = ["id", "label", "inputPorts", "outputPorts", "blockLibrary", "blockType", "properties"]
        missing = [field for field in required_fields if field not in data]
        if missing:
            raise ValueError(f"Missing required fields in HighLevelBlock: {', '.join(missing)}")

        raw_props = data["properties"]
        parsed_props = {}

        for key, entry in raw_props.items():
            ptype = entry.get("type")
            value = entry.get("value")

            # Evaluate only if it's a string and not a string-typed property
            if isinstance(value, str) and ptype != "string":
                try:
                    evaluated = eval(value, parameter_environment_namespace, parameter_environment_namespace)
                except Exception as e:
                    raise ValueError(f"Error evaluating property '{key}': {e}")
            else:
                evaluated = value

            parsed_props[key] = {
                "type": ptype,
                "value": evaluated
            }

        return cls(
            id=data["id"],
            label=data["label"],
            input_ports=data["inputPorts"],
            output_ports=data["outputPorts"],
            block_library=data["blockLibrary"],
            block_type=data["blockType"],
            properties=parsed_props,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "label": self.label,
            "inputPorts": self.input_ports,
            "outputPorts": self.output_ports,
            "blockLibrary": self.block_library,
            "blockType": self.block_type,
            "properties": self.properties,
        }


from enum import Enum
# ...existing code...
import json

class BlockShape(Enum):
    square = "square"
    triangle = "triangle"
    circle = "circle"

class FigurePath:
    x_values: list[float]
    y_values: list[float]
    color: str = '#000000'

    def to_dict(self):
        return {
            "x_values": self.x_values,
            "y_values": self.y_values,
            "color": self.color
        }

class BlockRenderFigure:
    paths: list[FigurePath] = []
    show_grid: bool = False
    show_axes: bool = True

    def to_dict(self):
        return {
            "paths": [p.to_dict() for p in self.paths],
            "show_grid": self.show_grid,
            "show_axes": self.show_axes
        }

class BlockRenderInformation:
    """
    If both figure and icon are set, figure has priority. If show_image_and_text is False,
    and either icon is not "" or figure is not None, icon or figure will be shown. If True,
    both the image and the text are shown.
    """
    shape: BlockShape = BlockShape.square
    icon: str = ""
    figure: BlockRenderFigure | None = None
    text: str = "No text"
    show_image_and_text: bool = False

    # New properties for dimensions
    default_width: float = 120.0
    default_height: float = 50.0
    min_width: float = 60.0
    min_height: float = 25.0
    max_width: float = 360.0
    max_height: float = 150.0

    input_ports: int = 1
    output_ports: int = 1

    def to_dict(self):
        return {
            "shape": self.shape.value,
            "icon": self.icon,
            "text": self.text,
            "show_image_and_text": self.show_image_and_text,
            "figure": self.figure.to_dict() if self.figure else None,
            "default_width": self.default_width,
            "default_height": self.default_height,
            "min_width": self.min_width,
            "min_height": self.min_height,
            "max_width": self.max_width,
            "max_height": self.max_height,
            "input_ports": self.input_ports,
            "output_ports": self.output_ports
        }

    def to_json(self):
        return json.dumps(self.to_dict())



import asyncio
import json
import os
import pathlib
import runpy
import traceback
import pysyslink_base
import yaml
from typing import Any, Callable, Dict, List

from pysyslink_toolkit.HighLevelBlock import HighLevelBlock
from pysyslink_toolkit.LowLevelBlockStructure import LowLevelBlockStructure
from pysyslink_toolkit.BlockRenderInformation import BlockRenderInformation
from pysyslink_toolkit.Plugin import BlockLibraryConfig
from pysyslink_toolkit.load_plugins import load_plugins_from_paths
from pysyslink_toolkit.compile_system import compile_pslk_to_yaml
from pysyslink_toolkit.simulate_system import simulate_system

def _load_config(config_path: str | None) -> Dict[str, Any]:
    if config_path is None:
        return {}
    with open(config_path, "r") as f:
        return yaml.safe_load(f)

def compile_system(config_path: str | None, high_level_system_path: str, output_yaml_path: str) -> str:
    """
    Compile a high-level system (dict) to a low-level system (dict).
    """

    try:
        compile_pslk_to_yaml(high_level_system_path, config_path, output_yaml_path)
        return 'success'
    except Exception as e:
        return 'failure: {}'.format(traceback.format_exc())

async def run_simulation(config_path: str | None, low_level_system: str, sim_options: str, 
                   display_callback: Callable[[pysyslink_base.ValueUpdateBlockEvent], None] = None) -> dict:
    """
    Run a simulation asynchronously (dummy implementation).
    """
    # This is a placeholder. You would implement your simulation logic here.
    # For now, just return a dummy result.

    config = _load_config(config_path)
    plugin_paths = config.get("base_block_type_support_plugin_paths", ["/usr/local/lib/pysyslink_plugins/block_type_supports"])
    plugin_configuration = config.get("base_plugin_configuration", {"BasicCppSupport/libraryPluginPath": "/usr/local/lib/pysyslink_plugins"})
    result = await simulate_system(
        low_level_system,
        sim_options,
        display_callback,
        plugin_paths,
        plugin_configuration
    )

    return result

def get_available_block_libraries(config_path: str | None) -> List[BlockLibraryConfig]:
    """
    Return all available libraries and blocks from loaded plugins.
    """
    plugins = load_plugins_from_paths(config_path)
    libraries: list[BlockLibraryConfig] = []
    for plugin in plugins:
        if hasattr(plugin, "get_block_libraries"):
            libraries.extend(plugin.get_block_libraries())
    return libraries

def get_block_render_information(config_path: str | None, block_data: Dict[str, Any], pslk_path: str) -> BlockRenderInformation:
    """
    Return render information for a block.
    """
    plugins = load_plugins_from_paths(config_path)

    with open(pslk_path, "r") as f:
        system_json = json.load(f)

    initialization_python_script_path = system_json.get("initialization_python_script_path", None)

    if initialization_python_script_path:
        # Resolve to absolute path if not already absolute
        if not os.path.isabs(initialization_python_script_path):
            pslk_dir = os.path.dirname(os.path.abspath(pslk_path))
            initialization_python_script_path = os.path.normpath(
                os.path.join(pslk_dir, initialization_python_script_path)
            )
        if os.path.isfile(initialization_python_script_path) and initialization_python_script_path.endswith(".py"):
            try:
                parameter_environment_dict = runpy.run_path(initialization_python_script_path, init_globals={})
            except Exception as e:
                raise RuntimeError(f"Initialization script {initialization_python_script_path} load failed") from e
        else:
            raise FileNotFoundError(f"Initialization script '{initialization_python_script_path}' not found or not a .py file.")
    else:
        print(f"No initialization script provided.")
        parameter_environment_dict = dict()

    block = HighLevelBlock.from_dict(block_data, parameter_environment_dict)
    for plugin in plugins:
        try:
            print(f"Testing plugin {plugin.name}")
            return plugin.get_block_render_information(block)
        except NotImplementedError:
            continue
        except Exception as e:
            raise RuntimeError(f"Exception while getting block render information: {e}")
    raise RuntimeError(f"No plugin could provide render information for block: {block.block_type}")

def get_block_html(config_path: str | None, block_data: Dict[str, Any], pslk_path: str) -> str:
    plugins = load_plugins_from_paths(config_path)

    with open(pslk_path, "r") as f:
        system_json = json.load(f)

    initialization_python_script_path = system_json.get("initialization_python_script_path", None)

    # Resolve to absolute path if not already absolute
    if not os.path.isabs(initialization_python_script_path):
        pslk_dir = os.path.dirname(os.path.abspath(pslk_path))
        initialization_python_script_path = os.path.normpath(
            os.path.join(pslk_dir, initialization_python_script_path)
        )

    if initialization_python_script_path:
        if os.path.isfile(initialization_python_script_path) and initialization_python_script_path.endswith(".py"):
            try:
                parameter_environment_dict = runpy.run_path(initialization_python_script_path, init_globals={})
            except Exception as e:
                raise RuntimeError(f"Initialization script {initialization_python_script_path} load failed") from e
        else:
            raise FileNotFoundError(f"Initialization script '{initialization_python_script_path}' not found or not a .py file.")
    else:
        raise FileNotFoundError(f"No initialization script provided.")

    block = HighLevelBlock.from_dict(block_data, parameter_environment_dict)
    for plugin in plugins:
        try:
            return plugin.get_block_html(block, pslk_path)
        except NotImplementedError:
            continue
        except Exception as e:
            raise RuntimeError(f"Exception while getting block render information: {e}")
    raise RuntimeError(f"No plugin could provide render information for block: {block.block_type}")

if __name__ == "__main__":
    test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "tests", "data")
    toolkit_config = os.path.join(test_dir, "toolkit_config.yaml")
    print(get_available_block_libraries(toolkit_config))


import importlib.util
import pathlib
import inspect
from typing import Any, Dict
from pysyslink_toolkit.CoreBlockPlugin import CoreBlockPlugin
from pysyslink_toolkit.Plugin import Plugin
import yaml

def _load_config(config_path: str | None) -> Dict[str, Any]:
    if config_path is None:
        return {}
    with open(config_path, "r") as f:
        return yaml.safe_load(f)

def load_plugins_from_paths(config_path, include_default_paths=True) -> list[Plugin]:
    plugins: list[Plugin] = []
    toolkit_config = _load_config(config_path)
    plugin_root_paths = toolkit_config.get("plugin_paths", [])
    print(plugin_root_paths)
    if include_default_paths:
        plugin_root_paths.append("/usr/local/lib/pysyslink_plugins")
    print(plugin_root_paths)
    for root in plugin_root_paths:
        root_path = pathlib.Path(root)
        if not root_path.is_absolute():
            root_path = pathlib.Path(config_path).parent / root_path
            root_path = root_path.resolve()
        for yaml_file in root_path.glob("**/*.pslkp.yaml"):
            with open(yaml_file, "r") as f:
                config = yaml.safe_load(f)
            if config["pluginType"] == "highLevelBlockLibrary":
                python_filename = config["pythonFilename"]
                py_path = yaml_file.parent / python_filename
                module_name = py_path.stem # py_path.stem is correct, it returns the module name
                try: 
                    plugins.append(load_plugin_from_file(py_path, module_name, config, toolkit_config))
                except ImportError as e:
                    print("Error loading plugin: {}".format(e.msg))
            elif config["pluginType"] == "coreBlockLibrary":
                plugins.append(CoreBlockPlugin(config, toolkit_config))

            
    return plugins

def load_plugin_from_file(path: pathlib.Path, module_name: str, yaml_config: dict, toolkit_config: dict) -> Plugin:

    spec = importlib.util.spec_from_file_location(module_name, str(path))
    if not spec or not spec.loader:
        raise ImportError(f"Cannot load module from {path}")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Find a subclass of Plugin in the module
    for name, obj in inspect.getmembers(module, inspect.isclass):
        if issubclass(obj, Plugin) and obj is not Plugin:
            return obj(yaml_config, toolkit_config)  # Instantiate and return the plugin

    raise ImportError(f"No subclass of Plugin found in {path}")




from typing import Any, Dict

import pysyslink_base
from pysyslink_toolkit.BlockRenderInformation import BlockRenderInformation
from pysyslink_toolkit.HighLevelBlock import HighLevelBlock
from pysyslink_toolkit.LowLevelBlockStructure import LowLevelBlock, LowLevelBlockStructure
from pysyslink_toolkit.Plugin import Plugin


class CoreBlockPlugin(Plugin):
    def __init__(self, plugin_yaml, toolkit_config: dict):
        super().__init__(plugin_yaml, toolkit_config)
        for block_library in self.block_libraries:
            block_library.name = "core_" + block_library.name
        print(self.config.metadata)
        self.block_type = self.config.blockType
    
    def _convert_property_types(self, block_library_name: str, block_type_name: str, properties: Dict[str, Dict[str, Any]]):
        # Find the block library and block type definition
        block_type = self.get_block_type_config(block_library_name, block_type_name)

        converted = {}
        for key, prop in properties.items():
            prop_configuration = block_type.configurationValues.get(key)
            value = prop["value"]
            if prop_configuration is None:
                converted[key] = value  # or raise error if you want strictness
                continue
            # Convert based on expected_type
            expected_type = prop_configuration.type
            try:
                if expected_type == "float":
                    converted[key] = float(value)
                elif expected_type == "int":
                    converted[key] = int(value)
                elif expected_type == "bool":
                    converted[key] = bool(value)
                elif expected_type == "string":
                    converted[key] = str(value)
                elif expected_type.endswith("[]") and isinstance(value, list):
                    base_type = expected_type[:-2]
                    if base_type == "float":
                        converted[key] = [float(v) for v in value]
                    elif base_type == "int":
                        converted[key] = [int(v) for v in value]
                    elif base_type == "string":
                        converted[key] = [str(v) for v in value]
                    else:
                        converted[key] = value
                else:
                    converted[key] = value
            except Exception as e:
                raise ValueError(f"Error converting type of property {key} to type {expected_type}, value {value}: {e}")
        return converted

    def _get_block_class(self, block_library, block_type) -> str:
        if block_library.startswith("core_"):
            block_library = block_library[5:]
        return block_library + "/" + block_type

    def _compile_block(self, high_level_block: HighLevelBlock) -> LowLevelBlockStructure:
        converted_properties = self._convert_property_types(
            high_level_block.block_library,
            high_level_block.block_type,
            high_level_block.properties
        )
        # Create a LowLevelBlock with the same attributes as the high-level block

        low_level_block = LowLevelBlock(
            id=high_level_block.id,
            name=high_level_block.label,
            block_type=self.block_type,
            block_class=self._get_block_class(high_level_block.block_library, high_level_block.block_type),
            **converted_properties
        )
        # Map input and output ports directly
        port_map = {}
        for i in range(high_level_block.input_ports):
            port_map[("input", i)] = (low_level_block.id, i)
        for i in range(high_level_block.output_ports):
            port_map[("output", i)] = (low_level_block.id, i)
        return LowLevelBlockStructure([low_level_block], [], port_map)

    def _get_block_render_information(self, high_level_block) -> BlockRenderInformation:
        render_information = BlockRenderInformation()
        render_information.text = high_level_block.block_type
        
        print("type config: {}".format(self.get_block_type_config(high_level_block.block_library, high_level_block.block_type)))
        render_information.shape = self.get_block_type_config(high_level_block.block_library, high_level_block.block_type).blockShape

        # --- Convert high-level block to low-level block first ---
        low_level_struct = self._compile_block(high_level_block)
        if not low_level_struct.blocks:
            raise ValueError("No low-level blocks generated from high-level block.")

        low_level_block = low_level_struct.blocks[0]
        block_configuration = low_level_block.to_dict()

        # --- Add/override required fields for base block instantiation ---
        block_configuration["BlockType"] = self.block_type
        block_configuration["BlockClass"] = self._get_block_class(high_level_block.block_library, high_level_block.block_type)
        block_configuration["Name"] = low_level_block.name
        block_configuration["Id"] = low_level_block.id

        base_block = self.get_base_block(block_configuration)
        render_information.input_ports = len(base_block.get_input_ports())
        render_information.output_ports = len(base_block.get_output_ports())
        return render_information

    
    def get_base_block(self, block_configuration: dict[str, Any]):
        try:
            pysyslink_base.SpdlogManager.configure_default_logger()
        except:
            pass
        
        pysyslink_base.SpdlogManager.set_log_level(pysyslink_base.LogLevel.off)

        plugin_dir = self.toolkit_config.get("base_block_type_support_plugin_paths", ["/usr/local/lib/pysyslink_plugins/block_type_supports"])
        plugin_configuration = self.toolkit_config.get("base_plugin_configuration", {"BasicCppSupport/libraryPluginPath": "/usr/local/lib/pysyslink_plugins"})

        plugin_loader = pysyslink_base.BlockTypeSupportPluginLoader()

        if isinstance(plugin_dir, str):
            plugin_dir = [plugin_dir]
        
        for dir in plugin_dir:
            block_factories = plugin_loader.load_plugins(dir, plugin_configuration)

        block_events_handler = pysyslink_base.BlockEventsHandler()

        print("Block factories: {}".format(block_factories))
        block_factory = block_factories.get(self.block_type, None)
        if block_factory != None:
            try:
                return block_factory.create_block(block_configuration, block_events_handler)
            except Exception as e:
                raise ValueError("Could not create a block of type {}. Exception: {}".format(self.block_type, e))
        else:
            raise ValueError("Block type: {} not found on block factories loaded as plugins on base.".format(self.block_type))
        



import abc
from typing import Any, Dict
import yaml

import pysyslink_base

from pysyslink_toolkit.HighLevelBlock import HighLevelBlock
from pysyslink_toolkit.BlockRenderInformation import BlockRenderInformation, BlockShape
from pysyslink_toolkit.LowLevelBlockStructure import LowLevelBlock, LowLevelBlockStructure


from dataclasses import dataclass, field
from typing import List, Optional, Union


@dataclass
class ConfigurationValue:
    name: str
    defaultValue: Union[float, int, str, List[float], List[int], None]
    type: str
    metadata: dict = field(default_factory=dict)

@dataclass
class BlockTypeConfig:
    name: str
    configurationValues: Dict[str, ConfigurationValue] = field(default_factory=dict)
    blockShape: BlockShape = BlockShape.square
    metadata: dict = field(default_factory=dict)

@dataclass
class BlockLibraryConfig:
    name: str
    blockTypes: List[BlockTypeConfig] = field(default_factory=list)
    metadata: dict = field(default_factory=dict)

@dataclass
class PluginConfig:
    pluginName: str
    pluginType: str
    blockType: str
    blockLibraries: List[BlockLibraryConfig] = field(default_factory=list)
    metadata: dict = field(default_factory=dict)

def dict_to_plugin_config(data: dict) -> PluginConfig:
    def parse_configuration_value(cv):
        known = {k: cv[k] for k in ['name', 'defaultValue', 'type'] if k in cv}
        metadata = {k: v for k, v in cv.items() if k not in known}
        return ConfigurationValue(**known, metadata=metadata)

    def parse_block_type(bt):
        known = {k: bt[k] for k in ['name', 'configurationValues', 'blockShape'] if k in bt}
        metadata = {k: v for k, v in bt.items() if k not in known}
        # Convert list of config values to dict by name
        config_values = {
            cv["name"]: parse_configuration_value(cv)
            for cv in bt.get("configurationValues", [])
        }
        return BlockTypeConfig(
            name=bt["name"],
            configurationValues=config_values,
            blockShape=BlockShape(bt.get("blockShape")) if "blockShape" in bt else BlockShape.square,
            metadata=metadata
        )

    def parse_block_library(bl):
        known = {k: bl[k] for k in ['name', 'blockTypes'] if k in bl}
        metadata = {k: v for k, v in bl.items() if k not in known}
        return BlockLibraryConfig(
            name=bl["name"],
            blockTypes=[parse_block_type(bt) for bt in bl.get("blockTypes", [])],
            metadata=metadata
        )

    known = {k: data[k] for k in ['pluginName', 'pluginType', 'blockType', 'blockLibraries'] if k in data}
    metadata = {k: v for k, v in data.items() if k not in known}
    return PluginConfig(
        pluginName=data.get("pluginName"),
        pluginType=data.get("pluginType"),
        blockType=data.get("blockType"),
        blockLibraries=[parse_block_library(bl) for bl in data.get("blockLibraries", [])],
        metadata=metadata
    )


class Plugin(abc.ABC):
    def __init__(self, plugin_yaml: dict, toolkit_config: dict):
        super().__init__()
        if isinstance(plugin_yaml, str):
            plugin_yaml = yaml.safe_load(plugin_yaml)
        elif isinstance(plugin_yaml, dict):
            pass
        else:
            raise ValueError("plugin_yaml must be a dict or YAML string")
        self.config = dict_to_plugin_config(plugin_yaml)
        self.name = self.config.pluginName
        self.plugin_type = self.config.pluginType
        self.block_libraries = self.config.blockLibraries
        self.toolkit_config = toolkit_config

    
    def get_block_type_config(self, block_library_name: str, block_type_name: str) -> Optional[BlockTypeConfig]:
        block_library = next(filter(lambda lib: lib.name == block_library_name, self.block_libraries), None)
        if block_library == None:
            raise NotImplementedError(f"Block library {block_library_name} not in plugin {self.name}")
        block_type = next(filter(lambda block_type: block_type.name == block_type_name, block_library.blockTypes))
        if block_type == None:
            raise NotImplementedError(f"Block type {block_type_name} not found on library {block_library.name} in plugin {self.name}")
        return block_type

    def compile_block(self, high_level_block: HighLevelBlock) -> LowLevelBlockStructure:
        block_type = self.get_block_type_config(high_level_block.block_library, high_level_block.block_type)
        return self._compile_block(high_level_block)

    @abc.abstractmethod
    def _compile_block(self, high_level_block: HighLevelBlock) -> LowLevelBlockStructure:
        pass

    def get_block_render_information(self, high_level_block: HighLevelBlock) -> BlockRenderInformation:
        block_type = self.get_block_type_config(high_level_block.block_library, high_level_block.block_type)
        print(f"Block type {block_type} obtained for high level block {high_level_block.block_library}, {high_level_block.block_type}")
        return self._get_block_render_information(high_level_block)

    @abc.abstractmethod
    def _get_block_render_information(self, high_level_block: HighLevelBlock) -> BlockRenderInformation:
        pass
    
    def get_block_html(self, high_level_block: HighLevelBlock, pslk_path: str) -> str:
        block_type = self.get_block_type_config(high_level_block.block_library, high_level_block.block_type)
        html_or_none = self._get_block_html(high_level_block, pslk_path)
        if html_or_none == None:
            return f"No HTML for block {high_level_block.label} of type {high_level_block.block_library} {high_level_block.block_type}"
        else:
            return html_or_none

    def _get_block_html(self, high_level_block: HighLevelBlock, pslk_path: str) -> str | None:
        return None

    def get_block_libraries(self):
        return self.block_libraries
    

    




from .BlockRenderInformation import BlockRenderInformation
from .HighLevelBlock import HighLevelBlock
from .LowLevelBlockStructure import LowLevelBlockStructure
from .Plugin import Plugin
from .api import *


import faulthandler
import re; faulthandler.enable()
import json
import time
import pysyslink_base
import yaml
from typing import Callable, Any    

async def simulate_system(
    system_yaml_path: str,
    sim_options_yaml_path: str,
    display_callback: Callable[[pysyslink_base.ValueUpdateBlockEvent], None] = None,
    plugin_dir: str | list[str] = "/usr/local/lib/pysyslink_plugins/block_type_supports/",
    plugin_configuration: dict[str, Any] = {"BasicCppSupport/libraryPluginPath": "/usr/local/lib/pysyslink_plugins"}
) -> dict:
    """
    Simulate a system using PySysLinkBase Python bindings.

    Args:
        system_yaml_path: Path to the system YAML file.
        sim_options_yaml_path: Path to the simulation options YAML file.
        output_filename: Path to save the simulation output YAML.
        display_callback: Function called as display is updated: (block_id, time, value).
        plugin_dir: Directory to search for block type support plugins.

    Returns:
        The simulation output object.
    """
    # Set up logging
    try:
        pysyslink_base.SpdlogManager.configure_default_logger()
    except:
        pass
    pysyslink_base.SpdlogManager.set_log_level(pysyslink_base.LogLevel.debug)

    # Block events handler (for display updates)
    block_events_handler = pysyslink_base.BlockEventsHandler()
    print(pysyslink_base.BlockEventsHandler.register_value_update_block_event_callback)
    block_events_handler.register_value_update_block_event_callback(display_callback)

    # Load plugins
    plugin_loader = pysyslink_base.BlockTypeSupportPluginLoader()

    if isinstance(plugin_dir, str):
        plugin_dir = [plugin_dir]
    
    for dir in plugin_dir:
        block_factories = plugin_loader.load_plugins(dir, plugin_configuration)
    
    print(system_yaml_path)


    # Parse the simulation model
    simulation_model = pysyslink_base.ModelParser.parse_from_yaml(
                        system_yaml_path, block_factories, block_events_handler
                    )

    block_chains = simulation_model.get_direct_block_chains()

    print("Block chains acquired")

    ordered_blocks = simulation_model.order_block_chains_onto_free_order(block_chains)

    print("Blocks ordered")

    # Propagate sample times
    simulation_model.propagate_sample_times()

    print("Simulation times propagated")

    # Load simulation options from YAML
    with open(sim_options_yaml_path, "r") as f:
        # This follows https://stackoverflow.com/questions/58434563/difference-between-yaml-load-and-yaml-safeloader-in-pyyaml to fix floating point on scientific notation
        loader = yaml.SafeLoader
        loader.add_implicit_resolver(
            u'tag:yaml.org,2002:float',
            re.compile(u'''^(?:
            [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$''', re.X),
            list(u'-+0123456789.'))
        sim_opts_dict = yaml.load(f, Loader=loader)

    simulation_options = pysyslink_base.SimulationOptions()
    simulation_options.start_time = sim_opts_dict.get("start_time", 0.0)
    simulation_options.stop_time = sim_opts_dict.get("stop_time", 10.0)
    simulation_options.run_in_natural_time = sim_opts_dict.get("run_in_natural_time", False)
    simulation_options.natural_time_speed_multiplier = sim_opts_dict.get("natural_time_speed_multiplier", 1)
    simulation_options.block_ids_input_or_output_and_indexes_to_log = sim_opts_dict.get(
        "block_ids_input_or_output_and_indexes_to_log", []
    )
    simulation_options.solvers_configuration = sim_opts_dict.get("solvers_configuration", {})

    output_filename = sim_opts_dict.get("simulation_output_filename")
    

    # Create and run simulation
    simulation_manager = pysyslink_base.SimulationManager(simulation_model, simulation_options)
    simulation_output = simulation_manager.run_simulation()

    # Save output to YAML
    with open(output_filename, "w") as f:
        all_signals = {}
        for signal_group, signals in simulation_output.signals.items():
            all_signals[signal_group] = {}
            for signal_name, signal_obj in signals.items():
                typed_signal = signal_obj.try_cast_to_typed()
                print("Signal group: {}, signal name: {}".format(signal_group, signal_name))
                all_signals[signal_group][signal_name] = {
                    "times": getattr(typed_signal, "times", []),
                    "values": getattr(typed_signal, "values", [])
                }
        json.dump(all_signals, f, indent=2)
        return all_signals


    



if __name__ == "__main__":
    import os

    # Use test data files for debugging
    test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "tests", "data")
    system_yaml_path = os.path.join(test_dir, "simulable_system.yaml")
    sim_options_yaml_path = os.path.join(test_dir, "sim_options.yaml")
    output_yaml_path = os.path.join(os.path.dirname(__file__), "..", "..", "tests", "test_outputs", "output_debug_simulation.yaml")

    os.makedirs(os.path.dirname(output_yaml_path), exist_ok=True)

    def debug_display_callback(event):
        print("[Python]:{},{},{}".format(event.value_id, event.simulation_time, event.value))

    simulate_system(
        system_yaml_path,
        sim_options_yaml_path,
        output_yaml_path,
        display_callback=debug_display_callback
    )
    print("outside")
